import logging
import sys
import time

from PIL import Image, ImageChops

from pyvirtualdisplay import Display

if sys.platform == "darwin" or sys.platform == "win32":
    from PIL.ImageGrab import grab
else:
    from pyscreenshot import grab

log = logging.getLogger(__name__)


class DisplayTimeoutError(Exception):
    pass


class SmartDisplay(Display):
    def autocrop(self, im):
        """Crop borders off an image.

        :param im: Source image.
        :param bgcolor: Background color, using either a color tuple or a color name (1.1.4 only).
        :return: An image without borders, or None if there's no actual content in the image.
        """
        if im.mode != "RGB":
            im = im.convert("RGB")
        bg = Image.new("RGB", im.size, self._bgcolor)
        diff = ImageChops.difference(im, bg)
        bbox = diff.getbbox()
        if bbox:
            return im.crop(bbox)
        return None  # no contents

    def grab(self, autocrop=True):
        img = grab()
        if sys.platform == "darwin" or sys.platform == "win32":
            img = grab(xdisplay=self.new_display_var)
        else:
            img = grab()

        if autocrop:
            img = self.autocrop(img)
        return img

    def waitgrab(self, timeout=60, autocrop=True, cb_imgcheck=None):
        """start process and create screenshot.
        Repeat screenshot until it is not empty and
        cb_imgcheck callback function returns True
        for current screenshot.

        :param autocrop: True -> crop screenshot
        :param timeout: int
        :param cb_imgcheck: None or callback for testing img,
                            True = accept img,
                            False = reject img
        """
        t = 0
        sleep_time = 0.3  # for fast windows
        repeat_time = 0.5
        while 1:
            log.debug("sleeping %s secs" % str(sleep_time))
            time.sleep(sleep_time)
            t += sleep_time
            img = self.grab(autocrop=autocrop)
            if img:
                if not cb_imgcheck:
                    break
                if cb_imgcheck(img):
                    break
            sleep_time = repeat_time
            repeat_time += 0.5  # progressive
            if t > timeout:
                msg = "Timeout! elapsed time:%s timeout:%s " % (t, timeout)
                raise DisplayTimeoutError(msg)
                # break

            log.debug("screenshot is empty, next try..")
        assert img
        #        if not img:
        #            log.debug('screenshot is empty!')
        return img



import random


class Randomizer:
    """
    Class to generate random display numbers between a minimum and maximum
    value. Note that the maximum value can be exceeded if every value between
    the generated and maximum value is already in use.
    """

    def __init__(self, min_display_nr=1000, max_display_nr=1100):
        self.min = min_display_nr
        self.delta = max_display_nr - min_display_nr

        random.seed()

    def generate(self):
        """
        Generate a random display number

        :rtype: int
        """

        return self.min + random.randint(0, self.delta)



"""Utility functions for xauth."""
import hashlib
import os

from easyprocess import EasyProcess


class NotFoundError(Exception):
    """Error when xauth was not found."""


def is_installed():
    """
    Return whether or not xauth is installed.
    """
    try:
        p = EasyProcess(["xauth", "-V"])
        p.enable_stdout_log = False
        p.enable_stderr_log = False
        p.call()
    except Exception:
        return False
    else:
        return True


def generate_mcookie():
    """
    Generate a cookie string suitable for xauth.
    """
    data = os.urandom(16)  # 16 bytes = 128 bit
    return hashlib.md5(data).hexdigest()


def call(*args):
    """
    Call xauth with the given args.
    """
    EasyProcess(["xauth"] + list(args)).call()



import sys

import pytest
from easyprocess import EasyProcess
from path import Path

from pyvirtualdisplay import Display
from pyvirtualdisplay.smartdisplay import DisplayTimeoutError, SmartDisplay

python = sys.executable


def test_disp():
    with Display():

        d = SmartDisplay(visible=True).start().stop()
        assert d.return_code == 0

        d = SmartDisplay(visible=False).start().stop()
        assert d.return_code == 0


def test_slowshot():
    disp = SmartDisplay(visible=False).start()
    py = Path(__file__).parent / ("slowgui.py")
    proc = EasyProcess([python, py]).start()
    img = disp.waitgrab()
    proc.stop()
    disp.stop()
    assert img is not None


def test_slowshot_with():
    disp = SmartDisplay(visible=False)
    py = Path(__file__).parent / ("slowgui.py")
    proc = EasyProcess([python, py])
    with disp:
        with proc:
            img = disp.waitgrab()
    assert img is not None


def test_empty():
    disp = SmartDisplay(visible=False)
    proc = EasyProcess([python])
    with disp:
        with proc:
            with pytest.raises(Exception):
                disp.waitgrab()


def test_slowshot_timeout():
    disp = SmartDisplay(visible=False)
    py = Path(__file__).parent / ("slowgui.py")
    proc = EasyProcess([python, py])
    with disp:
        with proc:
            with pytest.raises(DisplayTimeoutError):
                img = disp.waitgrab(timeout=1)



"Testing gnumeric on low resolution."
from easyprocess import EasyProcess

from pyvirtualdisplay import Display

# start Xephyr
with Display(visible=True, size=(320, 240)) as disp:
    # start Gnumeric
    with EasyProcess(["gnumeric"]) as proc:
        proc.wait()



import os

from pyvirtualdisplay import Display, xauth


def test_xauth():
    """
    Test that a Xauthority file is created.
    """
    if not xauth.is_installed():
        print("This test needs xauth installed")
        return
    old_xauth = os.getenv("XAUTHORITY")
    display = Display(visible=False, use_xauth=True)
    display.start()
    new_xauth = os.getenv("XAUTHORITY")

    assert new_xauth is not None
    assert os.path.isfile(new_xauth)
    filename = os.path.basename(new_xauth)
    assert filename.startswith("PyVirtualDisplay.")
    assert filename.endswith("Xauthority")

    display.stop()
    assert old_xauth == os.getenv("XAUTHORITY")
    assert not os.path.isfile(new_xauth)



import sys
from time import sleep

from easyprocess import EasyProcess
from entrypoint2 import entrypoint

from pyvirtualdisplay import Display
from tutil import platform_is_osx

# ubuntu 14.04 no displayfd
# ubuntu 16.04 displayfd
# ubuntu 18.04 displayfd


# TODO: osx error:            Cannot open "/tmp/server-0.xkm" to write keyboard description
if not platform_is_osx():

    def test_race_10_xvfb():
        check_N(10, "xvfb")

    def test_race_10_xephyr():
        check_N(10, "xephyr")

    def test_race_10_xvnc():
        check_N(10, "xvnc")


def check_N(N, backend):
    with Display():
        ls = []
        try:
            for i in range(N):
                cmd = [
                    sys.executable,
                    __file__.rsplit(".", 1)[0] + ".py",
                    str(i),
                    backend,
                    "--debug",
                ]
                p = EasyProcess(cmd)
                p.start()
                ls += [p]

            sleep(3)

            good_count = 0
            rc_ls = []
            for p in ls:
                p.wait()
                if p.return_code == 0:
                    good_count += 1
                rc_ls += [p.return_code]
        finally:
            for p in ls:
                p.stop()
        print(rc_ls)
        print(good_count)
        assert good_count == N


@entrypoint
def main(i, backend):
    # TODO: test all backends
    d = Display(backend=backend).start()
    print("my index:%s  backend:%s disp:%s" % (i, backend, d.new_display_var))
    ok = d.is_alive()
    d.stop()
    assert ok



"Start Xvfb server. Open xmessage window."

from easyprocess import EasyProcess

from pyvirtualdisplay import Display

with Display(visible=False, size=(100, 60)) as disp:
    with EasyProcess(["xmessage", "hello"]) as proc:
        proc.wait()



import logging
import os
import sys
from time import sleep

from easyprocess import EasyProcess
from pyscreenshot.tempdir import TemporaryDirectory

from pyvirtualdisplay import Display
from tutil import has_xvnc, prog_check

log = logging.getLogger(__name__)


python = sys.executable


def test_screenshot():
    owd = os.getcwd()
    with TemporaryDirectory(prefix="pyvirtualdisplay_") as tmpdirname:
        try:
            os.chdir(tmpdirname)
            p = EasyProcess([python, "-m", "pyvirtualdisplay.examples.screenshot"])
            p.call()
            assert p.return_code == 0
        finally:
            os.chdir(owd)


def test_headless():
    p = EasyProcess([python, "-m", "pyvirtualdisplay.examples.headless"]).start()
    sleep(1)
    assert p.is_alive()
    p.stop()


def test_nested():
    with Display():
        p = EasyProcess([python, "-m", "pyvirtualdisplay.examples.nested"]).start()
        sleep(1)
        assert p.is_alive()
        p.stop()


if has_xvnc():

    def test_vncserver():
        p = EasyProcess([python, "-m", "pyvirtualdisplay.examples.vncserver"]).start()
        sleep(1)
        assert p.is_alive()
        p.stop()


if prog_check(["gnumeric", "-help"]):

    def test_lowres():
        with Display():
            p = EasyProcess([python, "-m", "pyvirtualdisplay.examples.lowres"]).start()
            sleep(1)
            assert p.is_alive()
            p.stop()



import sys

from easyprocess import EasyProcess

from pyvirtualdisplay.util import get_helptext


def prog_check(cmd):
    try:
        return EasyProcess(cmd).call().return_code == 0
    except Exception:
        return False


def platform_is_osx():
    return sys.platform == "darwin"


def has_displayfd():
    return "-displayfd" in get_helptext("Xvfb")


def has_xvnc():
    return prog_check(["Xvnc", "-help"])



import logging

from pyvirtualdisplay.abstractdisplay import AbstractDisplay

log = logging.getLogger(__name__)

PROGRAM = "Xvnc"


class XvncDisplay(AbstractDisplay):
    """
    Xvnc wrapper
    """

    def __init__(
        self,
        size=(1024, 768),
        color_depth=24,
        bgcolor="black",
        use_xauth=False,
        # check_startup=False,
        rfbport=5900,
        rfbauth=None,
        randomizer=None,
        retries=10,
        extra_args=[],
    ):
        """
        :param bgcolor: 'black' or 'white'
        :param rfbport: Specifies the TCP port on which Xvnc listens for connections from viewers (the protocol used in VNC is called RFB - "remote framebuffer"). The default is 5900 plus the display number.
        :param rfbauth: Specifies the file containing the password used to authenticate viewers.
        """
        # self.screen = 0
        self._size = size
        self._color_depth = color_depth
        # self.process = None
        self._bgcolor = bgcolor
        # self.display = None
        self._rfbport = rfbport
        self._rfbauth = rfbauth

        AbstractDisplay.__init__(
            self,
            PROGRAM,
            use_xauth=use_xauth,
            # check_startup=check_startup,
            randomizer=randomizer,
            retries=retries,
            extra_args=extra_args,
        )

    def _check_flags(self, helptext):
        pass

    def _cmd(self):
        cmd = [
            PROGRAM,
            "-depth",
            str(self._color_depth),
            "-geometry",
            "%dx%d" % (self._size[0], self._size[1]),
            "-rfbport",
            str(self._rfbport),
        ]

        if self._rfbauth:
            cmd += ["-rfbauth", str(self._rfbauth)]
            # default:
            # -SecurityTypes = VncAuth
        else:
            cmd += ["-SecurityTypes", "None"]

        # cmd += [
        #     self.new_display_var,
        # ]

        # if self.check_startup:
        #     if self.has_displayfd:
        #         cmd += ["-displayfd", str(self.check_startup_fd)]
        if self._has_displayfd:
            cmd += ["-displayfd", str(self._pipe_wfd)]
        else:
            cmd += [self.new_display_var]
        return cmd



import fnmatch
import logging
import os
import select
import signal
import subprocess
import tempfile
import time
from threading import Lock

from easyprocess import EasyProcess, EasyProcessError

from pyvirtualdisplay import xauth
from pyvirtualdisplay.util import get_helptext, py2

# try:
#     import fcntl
# except ImportError:
#     fcntl = None


_mutex = Lock()

log = logging.getLogger(__name__)

_MIN_DISPLAY_NR = 1000
_USED_DISPLAY_NR_LIST = []

_X_START_TIMEOUT = 10
_X_START_TIME_STEP = 0.1
_X_START_WAIT = 0.1


class XStartTimeoutError(Exception):
    pass


class XStartError(Exception):
    pass


def _lock_files():
    tmpdir = "/tmp"
    try:
        ls = os.listdir(tmpdir)
    except FileNotFoundError:
        log.warning("missing /tmp")
        return []
    pattern = ".X*-lock"
    names = fnmatch.filter(ls, pattern)
    ls = [os.path.join(tmpdir, child) for child in names]
    ls = [p for p in ls if os.path.isfile(p)]
    return ls


def _search_for_display(randomizer=None):
    # search for free display
    ls = list(map(lambda x: int(x.split("X")[1].split("-")[0]), _lock_files()))
    if len(ls):
        display = max(_MIN_DISPLAY_NR, max(ls) + 3)
    else:
        display = _MIN_DISPLAY_NR

    if randomizer:
        display = randomizer.generate()

    return display


class AbstractDisplay(object):
    """
    Common parent for X servers (Xvfb,Xephyr,Xvnc)
    """

    def __init__(self, program, use_xauth, randomizer, retries, extra_args):
        self._extra_args = extra_args
        self._retries = retries
        self._program = program
        self._randomizer = randomizer
        self.stdout = None
        self.stderr = None
        self.old_display_var = None
        self._subproc = None
        self.display = None
        self.is_started = False

        helptext = get_helptext(program)
        self._has_displayfd = "-displayfd" in helptext
        if not self._has_displayfd:
            log.debug("-displayfd flag is missing.")
        # if check_startup and not has_displayfd:
        #     check_startup = False
        #     log.warning(
        #         program
        #         + " -displayfd flag is not supported, 'check_startup' parameter has been disabled"
        #     )
        self._check_flags(helptext)

        if use_xauth and not xauth.is_installed():
            raise xauth.NotFoundError()

        self._use_xauth = use_xauth
        self._old_xauth = None
        self._xauth_filename = None
        # self.check_startup = check_startup
        # if check_startup and not fcntl:
        #     self.check_startup = False
        #     log.warning(
        #         "fcntl module can't be imported, 'check_startup' parameter has been disabled"
        #     )
        #     log.warning("fnctl module does not exist on Windows")
        # if self.check_startup:
        #     rp, wp = os.pipe()
        #     fcntl.fcntl(rp, fcntl.F_SETFD, fcntl.FD_CLOEXEC)
        #     # TODO: to properly allow to inherit fds to subprocess on
        #     # python 3.2+ the easyprocess needs small fix..
        #     fcntl.fcntl(wp, fcntl.F_SETFD, 0)
        #     self.check_startup_fd = wp
        #     self._check_startup_fd = rp
        # self.proc = EasyProcess(self._cmd())

    def _check_flags(self, helptext):
        pass

    def _cmd(self):
        raise NotImplementedError()

    def _redirect_display(self, on):
        """
        on:
         * True -> set $DISPLAY to virtual screen
         * False -> set $DISPLAY to original screen

        :param on: bool
        """
        d = self.new_display_var if on else self.old_display_var
        if d is None:
            log.debug("unset $DISPLAY")
            try:
                del os.environ["DISPLAY"]
            except KeyError:
                log.warning("$DISPLAY was already unset.")
        else:
            log.debug("$DISPLAY=%s", d)
            os.environ["DISPLAY"] = d

    def start(self):
        """
        start display

        :rtype: self
        """
        if self.is_started:
            raise XStartError(self, "Display was started twice.")
        self.is_started = True

        if self._has_displayfd:
            self._start1()
        else:
            i = 0
            while True:
                try:
                    self._start1()
                    break
                except XStartError:
                    log.warning("start failed %s", i + 1)
                    time.sleep(0.05)
                    i += 1
                    if i >= self._retries:
                        raise XStartError(
                            "No success after %s retries. Last stderr: %s"
                            % (self._retries, self.stderr)
                        )
                finally:
                    self._redirect_display(False)
        self._redirect_display(True)

    def _start1(self):
        if self._has_displayfd:
            # stdout doesn't work on osx -> create own pipe
            rfd, self._pipe_wfd = os.pipe()
        else:
            with _mutex:
                self.display = _search_for_display(randomizer=self._randomizer)
                while self.display in _USED_DISPLAY_NR_LIST:
                    self.display += 1
                self.new_display_var = ":%s" % int(self.display)

                _USED_DISPLAY_NR_LIST.append(self.display)

        self._command = self._cmd() + self._extra_args
        log.debug("command: %s", self._command)

        self._stdout_file = tempfile.TemporaryFile(prefix="stdout_")
        self._stderr_file = tempfile.TemporaryFile(prefix="stderr_")

        if py2() or not self._has_displayfd:
            self._subproc = subprocess.Popen(
                self._command,
                stdout=self._stdout_file,
                stderr=self._stderr_file,
                shell=False,
            )
        else:
            if self._has_displayfd:
                self._subproc = subprocess.Popen(
                    self._command,
                    pass_fds=[self._pipe_wfd],
                    stdout=self._stdout_file,
                    stderr=self._stderr_file,
                    shell=False,
                )
        if self._has_displayfd:
            # rfd = self.subproc.stdout.fileno()
            self.display = int(self._wait_for_pipe_text(rfd))
            os.close(rfd)
            os.close(self._pipe_wfd)
        self.new_display_var = ":%s" % int(self.display)

        if self._use_xauth:
            self._setup_xauth()

        # https://github.com/ponty/PyVirtualDisplay/issues/2
        # https://github.com/ponty/PyVirtualDisplay/issues/14
        self.old_display_var = os.environ.get("DISPLAY", None)

        # wait until X server is active
        start_time = time.time()
        # if self.check_startup:
        #     rp = self._check_startup_fd
        #     display_check = None
        #     rlist, wlist, xlist = select.select((rp,), (), (), _X_START_TIMEOUT)
        #     if rlist:
        #         display_check = os.read(rp, 10).rstrip()
        #     else:
        #         msg = "No display number returned by X server"
        #         raise XStartTimeoutError(msg)
        #     dnbs = str(self.display)
        #     if bytes != str:
        #         dnbs = bytes(dnbs, "ascii")
        #     if display_check != dnbs:
        #         msg = 'Display number "%s" not returned by X server' + str(
        #             display_check
        #         )
        #         raise XStartTimeoutError(msg % self.display)

        if not self._has_displayfd:
            self._redirect_display(True)  # for xdpyinfo
            d = self.new_display_var
            ok = False
            time.sleep(0.05)  # give time for early exit
            while True:
                if not self.is_alive():
                    break

                try:
                    xdpyinfo = EasyProcess(["xdpyinfo"])
                    xdpyinfo.enable_stdout_log = False
                    xdpyinfo.enable_stderr_log = False
                    exit_code = xdpyinfo.call().return_code
                except EasyProcessError:
                    log.warning(
                        "xdpyinfo was not found, X start can not be checked! Please install xdpyinfo!"
                    )
                    time.sleep(_X_START_WAIT)  # old method
                    ok = True
                    break

                if exit_code != 0:
                    pass
                else:
                    log.info('Successfully started X with display "%s".', d)
                    ok = True
                    break

                if time.time() - start_time >= _X_START_TIMEOUT:
                    break
                time.sleep(_X_START_TIME_STEP)
            if not self.is_alive():
                log.warning("process exited early. stderr:%s", self.stderr)
                msg = "Failed to start process: %s"
                raise XStartError(msg % self)
            if not ok:
                msg = 'Failed to start X on display "%s" (xdpyinfo check failed, stderr:[%s]).'
                raise XStartTimeoutError(msg % (d, xdpyinfo.stderr))

        return self

    def _wait_for_pipe_text(self, rfd):
        s = ""
        start_time = time.time()
        while True:
            (rfd_changed_ls, _, _) = select.select([rfd], [], [], 0.1)
            if not self.is_alive():
                raise XStartError(
                    "%s program closed. command: %s stderr: %s"
                    % (self._program, self._command, self.stderr)
                )
            if rfd in rfd_changed_ls:
                c = os.read(rfd, 1)
                if c == b"\n":
                    break
                s += c.decode("ascii")
            if time.time() - start_time >= _X_START_TIMEOUT:
                raise XStartTimeoutError(
                    "No reply from program %s. command:%s"
                    % (self._program, self._command,)
                )
        return s

    def stop(self):
        """
        stop display

        :rtype: self
        """
        if not self.is_started:
            raise XStartError("stop() is called before start().")

        self._redirect_display(False)

        if self.is_alive():
            try:
                try:
                    self._subproc.terminate()
                except AttributeError:
                    os.kill(self._subproc.pid, signal.SIGKILL)
            except OSError as oserror:
                log.debug("exception in terminate:%s", oserror)

            self._subproc.wait()
            self._read_stdout_stderr()
        if self._use_xauth:
            self._clear_xauth()
        return self

    def _read_stdout_stderr(self):
        if self.stdout is None:
            self._stdout_file.seek(0)
            self._stderr_file.seek(0)
            self.stdout = self._stdout_file.read()
            self.stderr = self._stderr_file.read()

            self._stdout_file.close()
            self._stderr_file.close()

            log.debug("stdout=%s", self.stdout)
            log.debug("stderr=%s", self.stderr)

    def _setup_xauth(self):
        """
        Set up the Xauthority file and the XAUTHORITY environment variable.
        """
        handle, filename = tempfile.mkstemp(
            prefix="PyVirtualDisplay.", suffix=".Xauthority"
        )
        self._xauth_filename = filename
        os.close(handle)
        # Save old environment
        self._old_xauth = {}
        self._old_xauth["AUTHFILE"] = os.getenv("AUTHFILE")
        self._old_xauth["XAUTHORITY"] = os.getenv("XAUTHORITY")

        os.environ["AUTHFILE"] = os.environ["XAUTHORITY"] = filename
        cookie = xauth.generate_mcookie()
        xauth.call("add", self.new_display_var, ".", cookie)

    def _clear_xauth(self):
        """
        Clear the Xauthority file and restore the environment variables.
        """
        os.remove(self._xauth_filename)
        for varname in ["AUTHFILE", "XAUTHORITY"]:
            if self._old_xauth[varname] is None:
                del os.environ[varname]
            else:
                os.environ[varname] = self._old_xauth[varname]
        self._old_xauth = None

    def __enter__(self):
        """used by the :keyword:`with` statement"""
        self.start()
        return self

    def __exit__(self, *exc_info):
        """used by the :keyword:`with` statement"""
        self.stop()

    def is_alive(self):
        return self.return_code is None

    @property
    def return_code(self):
        rc = self._subproc.poll()
        if rc is not None:
            # proc exited
            self._read_stdout_stderr()
        return rc

    @property
    def pid(self):
        """
        PID (:attr:`subprocess.Popen.pid`)

        :rtype: int
        """
        if self._subproc:
            return self._subproc.pid



import logging

from pyvirtualdisplay.abstractdisplay import AbstractDisplay

log = logging.getLogger(__name__)

PROGRAM = "Xvfb"


class XvfbDisplay(AbstractDisplay):
    """
    Xvfb wrapper

    Xvfb is an X server that can run on machines with no display
    hardware and no physical input devices. It emulates a dumb
    framebuffer using virtual memory.
    """

    def __init__(
        self,
        size=(1024, 768),
        color_depth=24,
        bgcolor="black",
        use_xauth=False,
        # check_startup=False,
        fbdir=None,
        dpi=None,
        randomizer=None,
        retries=10,
        extra_args=[],
    ):
        """
        :param bgcolor: 'black' or 'white'
        :param fbdir: If non-null, the virtual screen is memory-mapped
            to a file in the given directory ('-fbdir' option)
        :param dpi: screen resolution in dots per inch if not None
        """
        self._screen = 0
        self._size = size
        self._color_depth = color_depth
        # self.process = None
        self._bgcolor = bgcolor
        # self.display = None
        self._fbdir = fbdir
        self._dpi = dpi

        AbstractDisplay.__init__(
            self,
            PROGRAM,
            use_xauth=use_xauth,
            # check_startup=check_startup,
            randomizer=randomizer,
            retries=retries,
            extra_args=extra_args,
        )

    def _check_flags(self, helptext):
        pass

    def _cmd(self):
        cmd = [
            dict(black="-br", white="-wr")[self._bgcolor],
            "-nolisten",
            "tcp",
            "-screen",
            str(self._screen),
            "x".join(map(str, list(self._size) + [self._color_depth])),
            # self.new_display_var,
        ]
        if self._fbdir:
            cmd += ["-fbdir", self._fbdir]
        if self._dpi is not None:
            cmd += ["-dpi", str(self._dpi)]
        # if self.check_startup:
        if self._has_displayfd:
            cmd += ["-displayfd", str(self._pipe_wfd)]
        else:
            cmd += [self.new_display_var]
        return [PROGRAM] + cmd



from pyvirtualdisplay.xephyr import XephyrDisplay
from pyvirtualdisplay.xvfb import XvfbDisplay
from pyvirtualdisplay.xvnc import XvncDisplay

_class_map = {"xvfb": XvfbDisplay, "xvnc": XvncDisplay, "xephyr": XephyrDisplay}


class Display(object):
    """
    Proxy class

    :param color_depth: [8, 16, 24, 32]
    :param size: screen size (width,height)
    :param bgcolor: background color ['black' or 'white']
    :param visible: True -> Xephyr, False -> Xvfb
    :param backend: 'xvfb', 'xvnc' or 'xephyr', ignores ``visible``
    :param xauth: If a Xauthority file should be created.
    """

    def __init__(
        self,
        backend=None,
        visible=False,
        size=(1024, 768),
        color_depth=24,
        bgcolor="black",
        use_xauth=False,
        # check_startup=False,
        randomizer=None,
        retries=10,
        extra_args=[],
        **kwargs
    ):
        self._color_depth = color_depth
        self._size = size
        self._bgcolor = bgcolor
        self._visible = visible
        self._backend = backend

        if not self._backend:
            if self._visible:
                self._backend = "xephyr"
            else:
                self._backend = "xvfb"

        cls = _class_map.get(self._backend)
        if not cls:
            raise ValueError("unknown backend: %s" % self._backend)

        self._obj = cls(
            size=size,
            color_depth=color_depth,
            bgcolor=bgcolor,
            randomizer=randomizer,
            use_xauth=use_xauth,
            # check_startup=check_startup,
            extra_args=extra_args,
            **kwargs
        )

    def start(self):
        """
        start display

        :rtype: self
        """
        self._obj.start()
        return self

    def stop(self):
        """
        stop display

        :rtype: self
        """
        self._obj.stop()
        return self

    def __enter__(self):
        """used by the :keyword:`with` statement"""
        self.start()
        return self

    def __exit__(self, *exc_info):
        """used by the :keyword:`with` statement"""
        self.stop()

    def is_alive(self):
        return self._obj.is_alive()

    @property
    def return_code(self):
        return self._obj.return_code

    @property
    def pid(self):
        """
        PID (:attr:`subprocess.Popen.pid`)

        :rtype: int
        """
        return self._obj.pid

    @property
    def is_started(self):
        return self._obj.is_started

    @property
    def display(self):
        return self._obj.display

    @property
    def new_display_var(self):
        return self._obj.new_display_var



from time import sleep

import pytest

from pyvirtualdisplay import Display
from pyvirtualdisplay.abstractdisplay import XStartError
from pyvirtualdisplay.randomize import Randomizer
from pyvirtualdisplay.xephyr import XephyrDisplay
from pyvirtualdisplay.xvfb import XvfbDisplay
from pyvirtualdisplay.xvnc import XvncDisplay
from tutil import has_xvnc


def test_virt():
    vd = Display().start().stop()
    assert vd.return_code == 0
    assert not vd.is_alive()


def test_random():
    r = Randomizer()
    vd = Display(randomizer=r).start().stop()
    assert vd.return_code == 0
    # assert r.min <= vd.display <= r.min + r.delta # can be auto nr
    assert not vd.is_alive()


def test_nest():
    vd = Display().start()
    assert vd.is_alive()

    nd = Display(visible=True).start().stop()

    assert nd.return_code == 0

    vd.stop()
    assert not vd.is_alive()


def test_disp():
    vd = Display().start()
    assert vd.is_alive()

    # d = Display(visible=True).start().sleep(2).stop()
    # .assertEquals(d.return_code, 0)

    d = Display(visible=False).start().stop()
    assert d.return_code == 0

    vd.stop()
    assert not vd.is_alive()


def test_repr_xvfb():
    display = Display()
    print(repr(display))

    display = Display(visible=False)
    print(repr(display))

    display = Display(backend="xvfb")
    print(repr(display))

    display = XvfbDisplay()
    print(repr(display))


if has_xvnc():

    def test_repr_xvnc():
        display = Display(backend="xvnc")
        print(repr(display))

        display = XvncDisplay()
        print(repr(display))


def test_repr_xephyr():
    display = Display(visible=True)
    print(repr(display))

    display = Display(backend="xephyr")
    print(repr(display))

    display = XephyrDisplay()
    print(repr(display))


def test_stop_nostart():
    with pytest.raises(XStartError):
        Display().stop()


def test_double_start():
    vd = Display()
    try:
        vd.start()
        with pytest.raises(XStartError):
            vd.start()
    finally:
        vd.stop()


def test_double_stop():
    vd = Display().start().stop()
    assert vd.return_code == 0
    assert not vd.is_alive()
    vd.stop()
    assert vd.return_code == 0
    assert not vd.is_alive()


def test_stop_terminated():
    vd = Display().start()
    assert vd.is_alive()
    vd._obj._subproc.terminate()
    sleep(0.2)
    assert not vd.is_alive()
    vd.stop()
    assert vd.return_code == 0
    assert not vd.is_alive()


def test_no_backend():
    with pytest.raises(ValueError):
        vd = Display(backend="unknown")


def test_color_xvfb():
    with pytest.raises(XStartError):
        vd = Display(color_depth=99).start().stop()
    vd = Display(color_depth=16).start().stop()
    vd = Display(color_depth=24).start().stop()
    vd = Display(color_depth=8).start().stop()


def test_color_xephyr():
    with Display():
        # requested screen depth not supported, setting to match hosts
        vd = Display(backend="xephyr", color_depth=99).start().stop()

        vd = Display(backend="xephyr", color_depth=16).start().stop()
        vd = Display(backend="xephyr", color_depth=24).start().stop()
        vd = Display(backend="xephyr", color_depth=8).start().stop()


if has_xvnc():

    def test_color_xvnc():
        with pytest.raises(XStartError):
            vd = Display(backend="xvnc", color_depth=99).start().stop()
        vd = Display(backend="xvnc", color_depth=16).start().stop()
        vd = Display(backend="xvnc", color_depth=24).start().stop()
        vd = Display(backend="xvnc", color_depth=8).start().stop()


def test_pid():
    with Display() as d:
        assert d.pid > 0
    with XvfbDisplay() as d:
        assert d.pid > 0


def test_bgcolor():
    Display(bgcolor="black").start().stop()
    Display(bgcolor="white").start().stop()
    with pytest.raises(KeyError):
        Display(bgcolor="green").start().stop()


def test_is_started():
    d = Display()
    assert not d.is_started
    d.start()
    assert d.is_started
    d.stop()
    assert d.is_started


def test_is_started2():
    with Display() as d:
        assert d.is_started
    assert d.is_started

    with XvfbDisplay() as d:
        assert d.is_started
    assert d.is_started

    with Display():
        with XephyrDisplay() as d:
            assert d.is_started
        assert d.is_started

        # with XvncDisplay() as d:
        #     assert d.is_started
        # assert d.is_started


def test_extra_args():
    # Unrecognized option
    d = Display(extra_args=["willcrash"])
    with pytest.raises(XStartError):
        d.start()

    with Display():
        # -c                     turns off key-click
        with Display(visible=True, extra_args=["-c"]) as d:
            assert d.is_alive()
        assert not d.is_alive()

        with XephyrDisplay(extra_args=["-c"]) as d:
            assert d.is_alive()
        assert not d.is_alive()


def test_display():
    d = Display()
    assert d.display is None
    d.start()
    assert d.display >= 0

    d = XvfbDisplay()
    assert d.display is None
    d.start()
    assert d.display >= 0



import os
from time import sleep

import fabric
import vagrant
from entrypoint2 import entrypoint
from path import Path

# pip3 install fabric vncdotool python-vagrant entrypoint2

DIR = Path(__file__).parent.parent.parent


class Options:
    halt = True
    recreate = True
    destroy = False


def run_box(options, vagrantfile, cmds):
    env = os.environ
    env["VAGRANT_VAGRANTFILE"] = DIR / vagrantfile
    if vagrantfile != "Vagrantfile":
        env["VAGRANT_DOTFILE_PATH"] = DIR / ".vagrant_" + vagrantfile
    else:
        env["VAGRANT_DOTFILE_PATH"] = ""

    v = vagrant.Vagrant(env=env, quiet_stdout=False, quiet_stderr=False)
    status = v.status()
    state = status[0].state
    print(status)

    if options.destroy:
        v.halt(force=True)
        v.destroy()
        return

    if options.halt:
        v.halt()  # avoid screensaver

    if state == "not_created":
        # install programs in box
        v.up()
        # restart box
        v.halt()

    try:
        v.up()

        with fabric.Connection(
            v.user_hostname_port(), connect_kwargs={"key_filename": v.keyfile(),},
        ) as conn:
            with conn.cd("c:/vagrant" if options.win else "/vagrant"):
                if not options.win:
                    cmds = ["env | sort"] + cmds
                sleep(1)
                for cmd in cmds:
                    if options.recreate:
                        if "tox" in cmd:
                            cmd += " -r"
                    # hangs without pty=True
                    conn.run(cmd, echo=True, pty=True)
    finally:
        if options.halt:
            v.halt()


config = {
    "server": ("Vagrantfile", ["tox"],),
    "server1404": ("Vagrantfile.14.04.rb", ["tox -e py27", "tox -e py36"],),
    "osx": ("Vagrantfile.osx.rb", ["bash --login -c 'python3 -m tox -e py3-osx'"],),
}


@entrypoint
def main(boxes="all", fast=False, destroy=False):
    options = Options()
    options.halt = not fast
    options.recreate = not fast
    options.destroy = destroy

    if boxes == "all":
        boxes = list(config.keys())
    else:
        boxes = boxes.split(",")
    for k, v in config.items():
        if k in boxes:
            options.win = k == "win"
            print("-----> %s %s %s" % (k, v[0], v[1]))
            run_box(options, v[0], v[1])



from backports import tempfile
from path import Path
from vncdotool import api

from pyvirtualdisplay import Display
from pyvirtualdisplay.xvnc import XvncDisplay
from tutil import has_xvnc

if has_xvnc():

    def test_color_xvnc():
        with tempfile.TemporaryDirectory() as temp_dir:
            vnc_png = Path(temp_dir) / "vnc.png"
            password = "123456"
            passwd_file = Path(temp_dir) / "pwd.txt"
            vncpasswd_generated = b"\x49\x40\x15\xf9\xa3\x5e\x8b\x22"
            passwd_file.write_bytes(vncpasswd_generated)

            with Display(backend="xvnc"):
                with api.connect("localhost:0") as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)
            with XvncDisplay():
                with api.connect("localhost:0") as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)

            with Display(backend="xvnc", rfbport=5900 + 9876):
                with api.connect("localhost:9876") as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)
            with XvncDisplay(rfbport=5900 + 9876):
                with api.connect("localhost:9876") as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)

            with Display(backend="xvnc", rfbauth=passwd_file):
                with api.connect("localhost:0", password=password) as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)
            with XvncDisplay(rfbauth=passwd_file):
                with api.connect("localhost:0", password=password) as client:
                    client.timeout = 1
                    client.captureScreen(vnc_png)



import sys

from easyprocess import EasyProcess


def get_helptext(program):
    p = EasyProcess([program, "-help"])
    p.enable_stdout_log = False
    p.enable_stderr_log = False
    p.call()
    helptext = p.stderr
    return helptext


def py2():
    return sys.version_info[0] == 2
